// 
// ChordFinderTest.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013, 2014 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using NUnit.Framework;

using com.jonthysell.Chordious.Core;

namespace com.jonthysell.Chordious.CoreTest
{
    [TestFixture()]
	public class ChordFinderTest
    {
        [SetUp()]
        public virtual void SetUp()
        {
            _instrument = new Instrument("Ukulele", 4);
            _instrument.AddTuning("Standard", new Note[] { Note.G, Note.C, Note.E, Note.A });

            foreach (Tuning t in _instrument.Tunings)
            {
                _tuning = t;
            }

            _majorChordQuality = new ChordQuality("Major", "", new int[] { 0, 4, 7 });
        }

        [TearDown()]
        public virtual void TearDown()
        {
            _chordFinder = null;
            _instrument = null;
            _tuning = null;
            _majorChordQuality = null;
            _chordFinderOptions = null;
        }

        #region New ChordFinder Via Params Tests

        [Test()]
        public void NewChordFinderViaParamsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);
            Assert.IsNotNull(_chordFinder);
        }

        [Test()]
        public void NewChordFinderViaParamsValidPropertiesTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);
            Assert.IsNotNull(_chordFinder);

            Assert.AreEqual(_instrument, _chordFinder.Instrument);
            Assert.AreEqual(_tuning, _chordFinder.Tuning);
        }

        #endregion

        #region FindChords Tests

        [Test()]
        public void FindChordsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {0, 0, 0, 3},
                new int[] {0, 4, 0, 3},
                new int[] {0, 4, 3, 3},
                new int[] {5, 4, 3, 3},
                new int[] {-1, 4, 3, 3},
                new int[] {0, 0, 0, -1},
                new int[] {5, 4, 3, -1},
                new int[] {0, -1, 0, 3},
                new int[] {0, 4, -1, 3},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, true, true, false, false);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        [Test()]
        public void FindChordsNoMutedStringsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {0, 0, 0, 3},
                new int[] {0, 4, 0, 3},
                new int[] {0, 4, 3, 3},
                new int[] {5, 4, 3, 3},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, true, false, false, false);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        [Test()]
        public void FindChordsNoOpenStringsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {5, 4, 3, 3},
                new int[] {-1, 4, 3, 3},
                new int[] {5, 4, 3, -1},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, false, true, false, false);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        [Test()]
        public void FindChordsNoMutedNoOpenTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {5, 4, 3, 3},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, false, false, false, false);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        [Test()]
        public void FindChordsAllowRootlessChordsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {0, 0, 0, 3},
                new int[] {0, 4, 0, 3},
                new int[] {0, 4, 3, 3},
                new int[] {5, 4, 3, 3},
                new int[] {-1, 4, 3, 3},
                new int[] {0, 0, 0, -1},
                new int[] {5, 4, 3, -1},
                new int[] {0, -1, 0, 3},
                new int[] {0, 4, -1, 3},
                new int[] {0, -1, 0, -1},
                new int[] {0, 4, 3, -1},
                new int[] {-1, 4, 3, -1},
                new int[] {0, 4, -1, -1},
                new int[] {0, 4, 0, -1},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, true, true, true, false);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        [Test()]
        public void FindChordsMirroredResultsTest()
        {
            _chordFinder = new ChordFinder(_instrument, _tuning);

            int[][] expectedChords = new int[][] {
                new int[] {3, 0, 0, 0},
                new int[] {3, 0, 4, 0},
                new int[] {3, 3, 4, 0},
                new int[] {3, 3, 4, 5},
                new int[] {3, 3, 4, -1},
                new int[] {-1, 0, 0, 0},
                new int[] {-1, 3, 4, 5},
                new int[] {3, 0, -1, 0},
                new int[] {3, -1, 4, 0},
            };

            _chordFinderOptions = new ChordFinderOptions(5, 5, 4, true, true, true, false, true);

            ChordResultSet chordResultSet = _chordFinder.FindChords(Note.C, _majorChordQuality, _chordFinderOptions);

            Assert.AreEqual(expectedChords.Length, chordResultSet.Count);

            foreach (int[] marks in expectedChords)
            {
                Assert.IsTrue(ChordResultSetContains(chordResultSet, marks));
            }
        }

        #endregion

        protected bool ChordResultSetContains(ChordResultSet chordResultSet, int[] expectedMarks)
        {
            for (int i = 0; i < chordResultSet.Count; i++)
            {
                int[] actualMarks = chordResultSet.ResultAt(i);

                if (expectedMarks.Length == actualMarks.Length)
                {
                    bool match = true;
                    for (int j = 0; j < expectedMarks.Length; j++)
                    {
                        match = match && (actualMarks[j] == expectedMarks[j]);
                    }

                    if (match)
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        protected ChordFinder _chordFinder;
        protected Instrument _instrument;
        protected Tuning _tuning;
        protected ChordQuality _majorChordQuality;
        protected ChordFinderOptions _chordFinderOptions;
    }
}