// 
// ChordTest.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using NUnit.Framework;

using com.jonthysell.Chordious.Core;

namespace com.jonthysell.Chordious.CoreTest
{
    [TestFixture()]
	public class ChordTest
    {
        [TearDown()]
        public virtual void TearDown()
        {
            _chord = null;
        }

        #region New Chord Via Params Tests

        [Test()]
        public void NewChordViaParamsTest()
        {
            _chord = new Chord(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);
            Assert.IsNotNull(_chord);
        }

        [Test()]
        public void NewChordViaParamsValidPropertiesTest()
        {
            _chord = new Chord(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);
            Assert.IsNotNull(_chord);

            Assert.AreEqual(_fileName, _chord.FileName);
            Assert.AreEqual(_title, _chord.Title);
            Assert.AreEqual(_numStrings, _chord.NumStrings);
            Assert.AreEqual(_numFrets, _chord.NumFrets);
            Assert.AreEqual(_baseLine, _chord.BaseLine);
            Assert.AreEqual(_barre, _chord.Barre);
        }

        #endregion

        #region New Chord Via ChordLine Tests

        [Test()]
        public void NewChordViaChordLineTest()
        {
            string chordLine = ToChordLine(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);
            _chord = new Chord(chordLine);
            Assert.IsNotNull(_chord);
        }

        [Test()]
        public void NewChordViaChordLineValidPropertiesTest()
        {
            string chordLine = ToChordLine(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);
            _chord = new Chord(chordLine);
            Assert.IsNotNull(_chord);

            Assert.AreEqual(_fileName, _chord.FileName);
            Assert.AreEqual(_title, _chord.Title);
            Assert.AreEqual(_numStrings, _chord.NumStrings);
            Assert.AreEqual(_numFrets, _chord.NumFrets);
            Assert.AreEqual(_baseLine, _chord.BaseLine);
            Assert.AreEqual(_barre, _chord.Barre);
        }

        [Test()]
        [ExpectedException(typeof(ArgumentNullException))]
        public void NewChordNullChordLineTest()
        {
            _chord = new Chord(null);
        }

        [Test()]
        [ExpectedException(typeof(ArgumentNullException))]
        public void NewChordEmptyChordLineTest()
        {
            _chord = new Chord("");
        }

        #endregion

        #region Property Tests

        [Test()]
        public void InvalidPropertyTests()
        {
            _chord = new Chord(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);

            int[] invalidValues = { 0, -1, Int32.MinValue };

            foreach (int iv in invalidValues)
            {
                Assert.Catch<ArgumentOutOfRangeException>(() => _chord.NumStrings =  iv);
                Assert.Catch<ArgumentOutOfRangeException>(() => _chord.NumFrets =  iv);
            }

            invalidValues = new int[]{ -1, Int32.MinValue };

            foreach (int iv in invalidValues)
            {
                Assert.Catch<ArgumentOutOfRangeException>(() => _chord.BaseLine =  iv);
            }

            Assert.Catch<ArgumentNullException>(() => _chord.Title =  null);
            Assert.Catch<ArgumentNullException>(() => _chord.Marks =  null);
        }

        [Test()]
        public void NullFileNameTests()
        {
            _chord = new Chord(null, _title, _numStrings, _numFrets, _baseLine, _barre, _marks);

            Assert.IsFalse(_chord.FileNameSet);
            Assert.IsNotNull(_chord.FileName);
            Assert.AreEqual(_chord.Title, _chord.FileName);

            _chord.FileName = _fileName;
            Assert.IsNotNull(_chord.FileName);
            Assert.IsTrue(_chord.FileNameSet);
            Assert.AreEqual(_fileName, _chord.FileName);

            _chord.FileName = null;
            Assert.IsNotNull(_chord.FileName);
            Assert.IsFalse(_chord.FileNameSet);
            Assert.AreEqual(_chord.Title, _chord.FileName);
        }

        [Test()]
        public void DifferentNumberOfMarksThanStringsTests()
        {
            int[] sameMarks = GenerateMarks(_numStrings);
            int[] lessMarks = GenerateMarks(_numStrings - 1);
            int[] moreMarks = GenerateMarks(_numStrings + 1);

            int[][] markSets = { sameMarks, lessMarks, moreMarks };

            foreach (int[] marks in markSets)
            {
                // At creation from parameters
                _chord = new Chord(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, marks);
                VerifyMarks(marks);
    
                // At creation from chordline
                string chordLine = ToChordLine(_fileName, _title, _numStrings, _numFrets, _baseLine, _barre, marks);
                _chord = new Chord(chordLine);
                VerifyMarks(marks);
    
                // At assignment
                _chord.Marks = marks;
                VerifyMarks(marks);
            }
        }

        #endregion

        protected string ToChordLine(string fileName, string title, int numStrings, int numFrets, int baseLine, int barre, int[] marks)
        {
            string s = String.Format("{0}:{1};{2};{3};{4};{5}", _fileName, _title, _numStrings, _numFrets, _baseLine, _barre);

            for (int i = 0; i < marks.Length; i++)
            {
                s+= ";" + marks[i];
            }

            return s;
        }

        protected int[] GenerateMarks(int length)
        {
            int[] range = new int[length];
            for (int i = 0; i < length; i++)
            {
                range[i] = i;
            }
            return range;
        }

        protected void VerifyMarks(int[] marks)
        {
            Assert.IsNotNull(_chord);
            Assert.AreEqual(_chord.NumStrings, _chord.Marks.Length);

            for (int i = 0; i < _chord.Marks.Length; i++)
            {
                if (i < marks.Length)
                {
                    Assert.AreEqual(marks[i], _chord.Marks[i]);
                }
                else
                {
                    Assert.AreEqual(0, _chord.Marks[i]);
                }
            }
        }

        protected Chord _chord;
        protected const string _title = "C";
        protected const string _fileName = "C";
        protected const int _numStrings = 4;
        protected const int _numFrets = 4;
        protected const int _baseLine = 0;
        protected const int _barre = -1;
        protected int[] _marks = { 0, 0, 0, 3 };
    }
}