// 
// MarkUtilsTest.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using NUnit.Framework;

using com.jonthysell.Chordious.Core;

namespace com.jonthysell.Chordious.CoreTest
{
    [TestFixture()]
	public class MarkUtilsTest
    {
        #region HasMutedStrings Tests

        [Test()]
        [ExpectedException(typeof(ArgumentNullException))]
        public void HasMutedStringsNullTest()
        {
            MarkUtils.HasMutedStrings(null);
        }

        [Test()]
        public void HasMutedStringsPositiveTest()
        {
            int[][] markSet = new int[][] {
                new int[] {-1},
                new int[] {-1, 0},
                new int[] {-1, 1},
                new int[] {0, -1},
                new int[] {1, -1},
                new int[] {0, -1, 0},
                new int[] {-1, 0, -1},
                new int[] {1, -1, 1},
                new int[] {-1, -1, -1},
                new int[] {-1, 1, -1},
            };

            foreach (int[] marks in markSet)
            {
                Assert.IsTrue(MarkUtils.HasMutedStrings(marks));
            }
        }

        [Test()]
        public void HasMutedStringsNegativeTest()
        {
            int[][] markSet = new int[][] {
                new int[] {0},
                new int[] {1},
                new int[] {0, 0},
                new int[] {1, 1},
                new int[] {0, 0, 0},
                new int[] {1, 1, 1},
            };

            foreach (int[] marks in markSet)
            {
                Assert.IsFalse(MarkUtils.HasMutedStrings(marks));
            }
        }

        #endregion

        #region HasOpenStrings Tests

        [Test()]
        [ExpectedException(typeof(ArgumentNullException))]
        public void HasOpenStringsNullTest()
        {
            MarkUtils.HasMutedStrings(null);
        }

        [Test()]
        public void HasOpenStringsPositiveTest()
        {
            int[][] markSet = new int[][] {
                new int[] {0},
                new int[] {-1, 0},
                new int[] {0, 1},
                new int[] {0, -1},
                new int[] {1, 0},
                new int[] {0, -1, 0},
                new int[] {-1, 0, -1},
                new int[] {1, 0, 1},
                new int[] {0, 0, 0},
                new int[] {0, 1, 0},
            };

            foreach (int[] marks in markSet)
            {
                Assert.IsTrue(MarkUtils.HasOpenStrings(marks));
            }
        }

        [Test()]
        public void HasOpenStringsNegativeTest()
        {
            int[][] markSet = new int[][] {
                new int[] {-1},
                new int[] {1},
                new int[] {-1,-10},
                new int[] {1, 1},
                new int[] {-1, -1, -1},
                new int[] {1, 1, 1},
            };

            foreach (int[] marks in markSet)
            {
                Assert.IsFalse(MarkUtils.HasOpenStrings(marks));
            }
        }

        #endregion

        #region AbsoluteToRelativeMarks Tests

        [Test()]
        [ExpectedException(typeof(ArgumentNullException))]
        public void AbsoluteToRelativeMarksNullTest()
        {
            int baseLine;
            MarkUtils.AbsoluteToRelativeMarks(null, out baseLine, _numFrets);
        }

        [Test()]
        public void AbsoluteToRelativeMarksReachExceedsFretsTest()
        {
            int[][] markSet = new int[][] {
                new int[] {-1, 0, 1},
                new int[] {1, 0, -1},
                new int[] {-1, 1, 1},
                new int[] {1, 1, -1},
                new int[] {-1, 1, 2},
                new int[] {2, 1, -1},
                new int[] {-1, 1, 3},
                new int[] {3, 1, -1},
            };

            int[] numFrets = {
                0,
                0,
                0,
                0,
                1,
                1,
                2,
                2,
            };

            for (int i = 0; i < markSet.Length; i++)
            {
                int baseLine;
                Assert.Catch<ArgumentOutOfRangeException>(() => MarkUtils.AbsoluteToRelativeMarks(markSet[i], out baseLine, numFrets[i]));
            }
        }

        [Test()]
        public void AbsoluteToRelativeMarksNoChangeTest()
        {
            int[][] markSet = new int[][] {
                new int[] {-1, 0, 1, _numFrets},
                new int[] {-1, 0, 1, _numFrets - 1},
                new int[] {_numFrets, 1, 0, -1},
                new int[] {_numFrets - 1, 1, 0, -1},
            };

            foreach (int[] marks in markSet)
            {
                int baseLine;
                int[] newMarks = MarkUtils.AbsoluteToRelativeMarks(marks, out baseLine, _numFrets);
                Assert.AreEqual(0, baseLine);
                Assert.IsTrue(CompareArrays(marks, newMarks));
            }
        }

        [Test()]
        public void AbsoluteToRelativeMarksBaseLineChangeTest()
        {
            int[][] markSet = new int[][] {
                new int[] {-1, 0, _numFrets + 1},
                new int[] {_numFrets + 1, 0, -1},
                new int[] {_numFrets + 1, _numFrets + 1, _numFrets + 1},
                new int[] {_numFrets + 1, _numFrets + 2, _numFrets + 3},
                new int[] {_numFrets + 3, _numFrets + 2, _numFrets + 1},
                new int[] {_numFrets + 2, _numFrets + 2, _numFrets + 2},
                new int[] {_numFrets + 2, _numFrets + 3, _numFrets + 4},
                new int[] {_numFrets + 4, _numFrets + 3, _numFrets + 2},
            };

            int[][] relativeMarkSet = new int[][] {
                new int[] {-1, 0, 1},
                new int[] {1, 0, -1},
                new int[] {1, 1, 1},
                new int[] {1, 2, 3},
                new int[] {3, 2, 1},
                new int[] {1, 1, 1},
                new int[] {1, 2, 3},
                new int[] {3, 2, 1},
            };

            int[] baseLines = {
                _numFrets + 1,
                _numFrets + 1,
                _numFrets + 1,
                _numFrets + 1,
                _numFrets + 1,
                _numFrets + 2,
                _numFrets + 2,
                _numFrets + 2,
            };

            for (int i = 0; i < markSet.Length; i++)
            {
                int baseLine;
                int[] newMarks = MarkUtils.AbsoluteToRelativeMarks(markSet[i], out baseLine, _numFrets);
                Assert.AreEqual(baseLines[i], baseLine);
                Assert.IsTrue(CompareArrays(relativeMarkSet[i], newMarks));
            }
        }

        protected bool CompareArrays(int[] a, int[] b)
        {
            if (ReferenceEquals(a, b))
            {
                return true;
            }

            if (a == null || b == null)
            {
                return false;
            }

            if (a.Length != b.Length)
            {
                return false;
            }

            for (int i = 0; i < a.Length; i++)
            {
                if (a[i] != b[i])
                {
                    return false;
                }
            }
            return true;
        }

        #endregion

        protected const int _numFrets = 4;
    }
}