// 
// ChordFinderDialog.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013, 2014 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

namespace com.jonthysell.Chordious.GtkUI
{
    public partial class ChordFinderDialog : Gtk.Dialog
    {
        public ChordFinderDialog(MainWindowVM mainWindowVm)
        {
            if (null == mainWindowVm)
            {
                throw new ArgumentNullException("mainWindowVm");
            }

            this.Build();

            _vm = new ChordFinderVM(mainWindowVm);

            Sensitive = false;
            LoadFromVM();
            Sensitive = true;
        }

        private void LoadFromVM()
        {
            // Load Top
            LoadInstruments();
            LoadTunings();
            LoadNotes();
            LoadQualities();

            // Load Side
            LoadOptions();
        }

        private void LoadInstruments()
        {
            FillComboBox(_instrumentComboBox, _vm.Instruments, _vm.ActiveInstrumentIndex);
        }

        private void LoadTunings()
        {
            FillComboBox(_tuningComboBox, _vm.Tunings, _vm.ActiveTuningIndex);
        }

        private void LoadNotes()
        {
            FillComboBox(_noteComboBox, _vm.RootNotes, _vm.ActiveRootNoteIndex);
        }

        private void LoadQualities()
        {
            FillComboBox(_qualityComboBox, _vm.Qualities, _vm.ActiveQualityIndex);
        }

        private void LoadOptions()
        {
            _numFretsSpinButton.Value = _vm.NumFrets;
            _maxFretSpinButton.Value = _vm.MaxFret;
            _maxReachSpinButton.Value = _vm.MaxReach;

            _autoAddBarresCheckButton.Active = _vm.AutoAddBarres;
            _allowOpenStringsCheckButton.Active = _vm.AllowOpenStrings;
            _allowMutedStringsCheckButton.Active = _vm.AllowMutedStrings;
            _allowRootlessChordsCheckButton.Active = _vm.AllowRootlessChords;
            _mirrorResultsCheckButton.Active = _vm.MirrorResults;

            AutoRefreshAction.Active = _autoRefresh;
        }

        private void FillComboBox(Gtk.ComboBox comboBox, List<string> items, int index)
        {
            comboBox.Sensitive = false;

            Gtk.ListStore store = new Gtk.ListStore(typeof (string));
            comboBox.Model = store;

            foreach (string item in items)
            {
                store.AppendValues(item);
            }

            comboBox.Active = index;

            comboBox.Sensitive = true;
        }

        private void LoadResults()
        {
            _resultsIconView.Sensitive = false;

            Gtk.ListStore store = new Gtk.ListStore(typeof (Gdk.Pixbuf));
            _resultsIconView.Model = store;
            _resultsIconView.PixbufColumn = 0;

            // Get new results

            foreach (string result in _vm.ResultSvgs)
            {
                Gdk.Pixbuf pixbuf = null;
                try
                {
                    pixbuf = GdkUtils.SvgToPixBuf(result);
                }
                catch (Exception)
                {
                    pixbuf = new Gdk.Pixbuf("icons/delete_2.png");
                }
                store.AppendValues(pixbuf);
            }

            // Unselect any previously selected items from a previous search
            _resultsIconView.UnselectAll();
            SaveSelectedAction.Sensitive = false;

            _resultsIconView.Sensitive = true;
        }

        private void PerformSearch()
        {
            try
            {
                Sensitive = false;
                _vm.FindChords();
                LoadResults();
            }
            catch (Exception) { }
            finally
            {
                Sensitive = true;
            }
        }

        private void SaveSelectedResults()
        {
            _resultsIconView.Sensitive = false;

            _vm.SaveSelectedResults();
            _vm.ClearSelectedResults();

            _resultsIconView.UnselectAll();
            SaveSelectedAction.Sensitive = false;

            _resultsIconView.Sensitive = true;
        }

        private void TryAutoRefresh()
        {
            if (_autoRefresh)
            {
                PerformSearch();
            }
        }
        private bool _autoRefresh = true;

        private ChordFinderVM _vm;

        protected void OnInstrumentComboBoxChanged(object sender, System.EventArgs e)
        {
            if (_instrumentComboBox.State == Gtk.StateType.Normal)
            {
                _vm.ActiveInstrumentIndex = _instrumentComboBox.Active;
                LoadTunings();
                TryAutoRefresh();
            }
        }

        protected void OnTuningComboBoxChanged(object sender, System.EventArgs e)
        {
            if (_tuningComboBox.State == Gtk.StateType.Normal)
            {
                _vm.ActiveTuningIndex = _tuningComboBox.Active;
                TryAutoRefresh();
            }
        }

        protected void OnNoteComboBoxChanged(object sender, System.EventArgs e)
        {
            if (_noteComboBox.State == Gtk.StateType.Normal)
            {
                _vm.ActiveRootNoteIndex = _noteComboBox.Active;
                TryAutoRefresh();
            }
        }

        protected void OnQualityComboBoxChanged(object sender, System.EventArgs e)
        {
            if (_qualityComboBox.State == Gtk.StateType.Normal)
            {
                _vm.ActiveQualityIndex = _qualityComboBox.Active;
                TryAutoRefresh();
            }
        }

        protected void OnNumFretsSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_numFretsSpinButton.State == Gtk.StateType.Normal)
            {
                _vm.NumFrets = _numFretsSpinButton.ValueAsInt;
                TryAutoRefresh();
            }
        }

        protected void OnMaxFretSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_maxFretSpinButton.State == Gtk.StateType.Normal)
            {
                _vm.MaxFret = _maxFretSpinButton.ValueAsInt;
                TryAutoRefresh();
            }
        }

        protected void OnMaxReachSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_maxReachSpinButton.State == Gtk.StateType.Normal)
            {
                _vm.MaxReach = _maxReachSpinButton.ValueAsInt;
                TryAutoRefresh();
            }
        }

        protected void OnAutoAddBarresCheckButtonToggled(object sender, System.EventArgs e)
        {
            if (_autoAddBarresCheckButton.State != Gtk.StateType.Insensitive)
            {
                _vm.AutoAddBarres = _autoAddBarresCheckButton.Active;
                TryAutoRefresh();
            }
        }

        protected void OnAllowOpenStringsCheckButtonToggled(object sender, System.EventArgs e)
        {
            if (_allowOpenStringsCheckButton.State != Gtk.StateType.Insensitive)
            {
                _vm.AllowOpenStrings = _allowOpenStringsCheckButton.Active;
                TryAutoRefresh();
            }
        }

        protected void OnAllowMutedStringsCheckButtonToggled(object sender, System.EventArgs e)
        {
            if (_allowMutedStringsCheckButton.State != Gtk.StateType.Insensitive)
            {
                _vm.AllowMutedStrings = _allowMutedStringsCheckButton.Active;
                TryAutoRefresh();
            }
        }

        protected void OnAllowRootlessChordsCheckButtonToggled(object sender, System.EventArgs e)
        {
            if (_allowRootlessChordsCheckButton.State != Gtk.StateType.Insensitive)
            {
                _vm.AllowRootlessChords = _allowRootlessChordsCheckButton.Active;
                TryAutoRefresh();
            }
        }

        protected void OnMirrorResultsCheckButtonToggled(object sender, System.EventArgs e)
        {
            if (_mirrorResultsCheckButton.State != Gtk.StateType.Insensitive)
            {
                _vm.MirrorResults = _mirrorResultsCheckButton.Active;
                TryAutoRefresh();
            }
        }

        protected void OnFindChordsActionActivated(object sender, System.EventArgs e)
        {
            PerformSearch();
        }

        protected void OnSaveSelectedActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                Sensitive = false;
                SaveSelectedResults();
            }
            catch (Exception) { }
            finally
            {
                Sensitive = true;
            }
        }

        protected void OnAutoRefreshActionToggled(object sender, System.EventArgs e)
        {
            try
            {
                Sensitive = false;
                _autoRefresh = AutoRefreshAction.Active;
                TryAutoRefresh();
            }
            catch (Exception) { }
            finally
            {
                Sensitive = true;
            }
        }

        protected void OnResultsIconViewSelectionChanged(object sender, System.EventArgs e)
        {
            try
            {
                _resultsIconView.Sensitive = false;
                _vm.ClearSelectedResults();
                foreach (Gtk.TreePath treePath in _resultsIconView.SelectedItems)
                {
                    _vm.SelectResult(treePath.Indices[0], true);
                }
            }
            catch (Exception) { }
            finally
            {
                SaveSelectedAction.Sensitive = _vm.SelectedResultsCount > 0;
                _resultsIconView.Sensitive = true;
            }
        }
    }
}