// 
// ChordFinderVM.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013, 2014 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using com.jonthysell.Chordious.Core;

namespace com.jonthysell.Chordious.GtkUI
{
    public class ChordFinderVM
    {
        public ChordFinderVM(MainWindowVM mainWindowVm)
        {
            if (null == mainWindowVm)
            {
                throw new ArgumentNullException("mainWindowVm");
            }

            this._mainWindowVm = mainWindowVm;
            Initialize();
        }

        private void Initialize()
        {
            _chordFinderOptions = new ChordFinderOptions();
            _chordOptions = _mainWindowVm.ApplicableChordOptions;

            LoadInstruments();
            LoadQualities();
            InitResults();
        }

        #region Instrument Selection

        public void LoadInstruments()
        {
            _instruments = new List<Instrument>(_mainWindowVm.ConfigFile.Instruments.Values);
            _activeInstrumentIndex = _instruments.Count > 0 ? 0 : -1;
            LoadTunings();
        }
        private List<Instrument> _instruments;

        public List<string> Instruments
        {
            get
            {
                List<string> instrumentList = new List<string>();

                foreach (Instrument instrument in _instruments)
                {
                    instrumentList.Add(instrument.Name);
                }

                return instrumentList;
            }
        }

        public int ActiveInstrumentIndex
        {
            get
            {
                return _activeInstrumentIndex;
            }
            set
            {
                if (value < 0 || value >= _instruments.Count)
                {
                    _activeInstrumentIndex = -1;
                }
                else
                {
                    _activeInstrumentIndex = value;
                }
                LoadTunings();
            }
        }
        private int _activeInstrumentIndex;

        private Instrument ActiveInstrument
        {
            get
            {
                try
                {
                    return _instruments[_activeInstrumentIndex];
                }
                catch (Exception) { }

                return null;
            }
        }

        #endregion

        #region Tuning Selection

        public void LoadTunings()
        {
            _tunings = new List<Tuning>();

            if (null != ActiveInstrument)
            {
                _tunings.AddRange(ActiveInstrument.Tunings);
            }

            _activeTuningIndex = _tunings.Count > 0 ? 0 : -1;
        }
        private List<Tuning> _tunings;

        public List<string> Tunings
        {
            get
            {
                List<string> tuningList = new List<string>();

                foreach (Tuning tuning in _tunings)
                {
                    tuningList.Add(tuning.LongName);
                }

                return tuningList;
            }
        }

        public int ActiveTuningIndex
        {
            get
            {
                return _activeTuningIndex;
            }
            set
            {
                if (value < 0 || value >= _tunings.Count)
                {
                    _activeTuningIndex = -1;
                }
                else
                {
                    _activeTuningIndex = value;
                }
            }
        }
        private int _activeTuningIndex;

        private Tuning ActiveTuning
        {
            get
            {
                try
                {
                    if (null != ActiveInstrument)
                    {
                        return _tunings[_activeTuningIndex];
                    }
                }
                catch (Exception) { }

                return null;
            }
        }

        #endregion

        #region Root Note Selection

        public List<string> RootNotes
        {
            get
            {
                List<string> noteList = new List<string>();

                foreach (Note note in Enum.GetValues(typeof(Note)))
                {
                    noteList.Add(NoteUtils.ToString(note));
                }

                return noteList;
            }
        }

        public int ActiveRootNoteIndex { get; set; }

        private Note ActiveRootNote
        {
            get
            {
                return (Note)ActiveRootNoteIndex;
            }
        }

        #endregion

        #region Quality Selection

        public void LoadQualities()
        {
            _qualities = new List<ChordQuality>(_mainWindowVm.ConfigFile.ChordQualities.Values);
            _activeQualityIndex = _qualities.Count > 0 ? 0 : -1;
        }
        private List<ChordQuality> _qualities;

        public List<string> Qualities
        {
            get
            {
                List<string> qualityList = new List<string>();

                foreach (ChordQuality quality in _qualities)
                {
                    qualityList.Add(quality.Name);
                }

                return qualityList;
            }
        }

        public int ActiveQualityIndex
        {
            get
            {
                return _activeQualityIndex;
            }
            set
            {
                if (value < 0 || value >= _qualities.Count)
                {
                    _activeQualityIndex = -1;
                }
                else
                {
                    _activeQualityIndex = value;
                }
            }
        }
        private int _activeQualityIndex;

        private ChordQuality ActiveQuality
        {
            get
            {
                try
                {
                    return _qualities[_activeQualityIndex];
                }
                catch (Exception) { }

                return null;
            }
        }

        #endregion

        #region Chord Finder Options

        public int NumFrets
        {
            get
            {
                return _chordFinderOptions.NumFrets;
            }
            set
            {
                _chordFinderOptions.NumFrets = value;
            }
        }

        public int MaxFret
        {
            get
            {
                return _chordFinderOptions.MaxFret;
            }
            set
            {
                _chordFinderOptions.MaxFret = value;
            }
        }

        public int MaxReach
        {
            get
            {
                return _chordFinderOptions.MaxReach;
            }
            set
            {
                _chordFinderOptions.MaxReach = value;
            }
        }

        public bool AutoAddBarres
        {
            get
            {
                return _chordFinderOptions.AutoAddBarres;
            }
            set
            {
                _chordFinderOptions.AutoAddBarres = value;
            }
        }

        public bool AllowOpenStrings
        {
            get
            {
                return _chordFinderOptions.AllowOpenStrings;
            }
            set
            {
                _chordFinderOptions.AllowOpenStrings = value;
            }
        }

        public bool AllowMutedStrings
        {
            get
            {
                return _chordFinderOptions.AllowMutedStrings;
            }
            set
            {
                _chordFinderOptions.AllowMutedStrings = value;
            }
        }

        public bool AllowRootlessChords
        {
            get
            {
                return _chordFinderOptions.AllowRootlessChords;
            }
            set
            {
                _chordFinderOptions.AllowRootlessChords = value;
            }
        }

        public bool MirrorResults
        {
            get
            {
                return _chordFinderOptions.MirrorResults;
            }
            set
            {
                _chordFinderOptions.MirrorResults = value;
            }
        }

        private ChordFinderOptions _chordFinderOptions;

        #endregion

        #region Chord Finding

        public void FindChords()
        {
            ChordFinder cf = new ChordFinder(ActiveInstrument, ActiveTuning);

            ChordResultSet crs = cf.FindChords(ActiveRootNote, ActiveQuality, _chordFinderOptions);

            InitResults();

            for (int i = 0; i < crs.Count; i++)
            {
                _chords.Add(crs.ChordAt(i));
            }

            if (_chords.Count > 0)
            {
                _selectedChords = new bool[_chords.Count];
            }
        }

        private List<Chord> _chords;
        private bool[] _selectedChords;

        #endregion

        #region Result Management

        public void InitResults()
        {
            _chords = new List<Chord>();
            _selectedChords = null;
        }

        public void ClearSelectedResults()
        {
            _selectedChords = null;
            if (null != _chords && _chords.Count > 0)
            {
                _selectedChords = new bool[_chords.Count];
            }
        }

        public void SelectResult(int index, bool selected)
        {
            if (null != _selectedChords)
            {
                _selectedChords[index] = selected;
            }
        }

        public void SaveSelectedResults()
        {
            if (null != _selectedChords)
            {
                for (int i = 0; i < _selectedChords.Length; i++)
                {
                    if (_selectedChords[i])
                    {
                        _mainWindowVm.AddChord(_chords[i]);
                    }
                }
            }
        }

        public int SelectedResultsCount
        {
            get
            {
                int count = 0;
                if (null != _selectedChords)
                {
                    for (int i = 0; i < _selectedChords.Length; i++)
                    {
                        if (_selectedChords[i])
                        {
                            count++;
                        }
                    }
                }
                return count;
            }
        }

        public List<string> ResultSvgs
        {
            get
            {
                List<string> svgList = new List<string>();
    
                foreach (Chord chord in _chords)
                {
                    svgList.Add(chord.ToSvg(_chordOptions));
                }
    
                return svgList;
            }
        }

        private ChordOptions _chordOptions;

        #endregion

        private MainWindowVM _mainWindowVm;
    }
}