// 
// GdkUtils.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using Gdk;

namespace com.jonthysell.Chordious.GtkUI
{
    public class GdkUtils
    {
        public static Pixbuf SvgToPixBuf(string svgText)
        {
            MemoryStream svgStream = new MemoryStream();
            StreamWriter writer = new StreamWriter(svgStream);
            writer.Write(svgText);
            writer.Flush();
            svgStream.Position = 0;

            return Rsvg.Pixbuf.LoadFromStream(svgStream);
        }

        public static void SvgToPngWriter(string svgText, string fileName)
        {
            Pixbuf pixbuf = SvgToPixBuf(svgText);
            fileName = Path.ChangeExtension(fileName, "png");
            pixbuf.Save(fileName, "png");
        }

        public static void SvgToJpegWriter(string svgText, string fileName)
        {
            Pixbuf pixbuf = SvgToPixBuf(svgText);

            // Need to remove transparency
            pixbuf = RemoveAlpha(pixbuf, WHITE);

            fileName = Path.ChangeExtension(fileName, "jpg");
            pixbuf.Save(fileName, "jpeg");
        }

        public static Pixbuf RemoveAlpha(Pixbuf pixbuf, byte[] background)
        {
            if (null == pixbuf)
            {
                throw new ArgumentNullException("pixbuf");
            }

            if (null == background)
            {
                throw new ArgumentNullException("background");
            }

            if (background.Length != 3)
            {
                throw new ArgumentException();
            }

            if (pixbuf.HasAlpha && pixbuf.NChannels == 4 && pixbuf.BitsPerSample == 8)
            {
                int width = pixbuf.Width;
                int height = pixbuf.Height;
                int rowStride = pixbuf.Rowstride;

                int origSize = width * pixbuf.NChannels * height;
                byte[] origPixels = new byte[origSize];
                System.Runtime.InteropServices.Marshal.Copy(pixbuf.Pixels, origPixels, 0, origSize);

                byte[] pixels = new byte[rowStride * height];

                for (int i = 0; i < height; i++) {
                    for (int j = 0, k = 0; j < width * 4; j += 4, k += 3) {
                        int origIndex = rowStride * i + j;
                        int index = rowStride * i + k;

                        double sourceR = Convert.ToDouble(origPixels[origIndex]) / Convert.ToDouble(Byte.MaxValue);
                        double sourceG = Convert.ToDouble(origPixels[origIndex + 1]) / Convert.ToDouble(Byte.MaxValue);
                        double sourceB = Convert.ToDouble(origPixels[origIndex + 2]) / Convert.ToDouble(Byte.MaxValue);
                        double sourceA = Convert.ToDouble(origPixels[origIndex + 3]) / Convert.ToDouble(Byte.MaxValue);

                        double backgroundR = Convert.ToDouble(background[0]) / Convert.ToDouble(Byte.MaxValue);
                        double backgroundG = Convert.ToDouble(background[1]) / Convert.ToDouble(Byte.MaxValue);
                        double backgroundB = Convert.ToDouble(background[2]) / Convert.ToDouble(Byte.MaxValue);

                        double alpha = sourceA;
                        double inverseAlpha = (1.0 - sourceA);

                        double targetR = Math.Min(1.0, (alpha * sourceR) + (inverseAlpha * backgroundR));
                        double targetG = Math.Min(1.0, (alpha * sourceG) + (inverseAlpha * backgroundG));
                        double targetB = Math.Min(1.0, (alpha * sourceB) + (inverseAlpha * backgroundB));

                        pixels[index] = Convert.ToByte(Math.Round(targetR * Byte.MaxValue));
                        pixels[index + 1] = Convert.ToByte(Math.Round(targetG * Byte.MaxValue));
                        pixels[index + 2] = Convert.ToByte(Math.Round(targetB * Byte.MaxValue));
                    }
                }

                return new Pixbuf(pixels, pixbuf.Colorspace, false, pixbuf.BitsPerSample, width, height, rowStride, null);
            }

            return pixbuf;
        }

        private static byte[] WHITE = new byte[] { 0xFF, 0xFF, 0xFF };
    }
}