// 
// MainWindow.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013, 2014 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using Gtk;

namespace com.jonthysell.Chordious.GtkUI
{
    public partial class MainWindow: Gtk.Window
    {	
        public MainWindow(): base(Gtk.WindowType.Toplevel)
        {
            Build();
            NewChordDocument();
        }

        private MainWindowVM _vm;

        public void NewChordDocument()
        {
            try
            {
                Sensitive = false;
                _vm = new MainWindowVM();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        public void LoadChordDocument(string inputFile)
        {
            try
            {
                Sensitive = false;
                _vm.LoadChordDocument(inputFile);
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        public bool TrySave()
        {
            try
            {
                TrySaveDirtyLine();
                _vm.SaveChordDocument();
                return true;
            }
            catch (FileNotFoundException)
            {
                return TrySaveAs();
            }
        }

        public bool TrySaveAs()
        {
            TrySaveDirtyLine();
            FileChooserDialog dialog = new FileChooserDialog("Save As...", this, FileChooserAction.Save,
                    "Cancel", ResponseType.Cancel, "Save", ResponseType.Accept);

            ResponseType result = (ResponseType)dialog.Run();
            string fileName = dialog.Filename;
            dialog.Destroy();

            if (result == ResponseType.Accept)
            {
                _vm.SaveChordDocumentAs(fileName);
                return true;
            }

            return false;
        }

        public bool TryClose()
        {
            TrySaveDirtyLine();
            if (_vm.DocumentIsDirty)
            {
                string fileName = _vm.DocumentFileName;
                MessageDialog dialog = new MessageDialog(this, DialogFlags.Modal, MessageType.Warning, ButtonsType.None, String.Format("Save the changes to document '{0}' before closing? If you don't save, all changes will be permanently lost.", fileName));

                dialog.AddButton("Close without Saving", ResponseType.No);
                dialog.AddButton("Cancel", ResponseType.Cancel);
                dialog.AddButton("Save", ResponseType.Yes);

                ResponseType result = (ResponseType)dialog.Run();
                dialog.Destroy();

                if (result == ResponseType.Yes)
                {
                    if (TrySave())
                    {
                        return TryClose();
                    }
                }
                else if (result == ResponseType.No)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            return true;
        }

        #region UI Refreshes

        protected void RefreshChordLines()
        {
            // If chord lines need refreshing, probably the title does too
            Title = _vm.WindowTitle;

            EnableChordLinesUI(false);

            // Update list of lines

            if (_chordLinesTreeView.Columns.Length == 0)
            {
                _chordLinesTreeView.AppendColumn("Lines", new Gtk.CellRendererText(), "text", 0);
            }

            Gtk.ListStore lineListStore = new Gtk.ListStore(typeof (string));

            foreach (string line in _vm.ChordLines)
            {
                lineListStore.AppendValues(line);
            }

            _chordLinesTreeView.Model = lineListStore;

            // Update selected item
            if (_vm.ActiveLineItem < 0)
            {
                RefreshChordLineActions();
                RefreshChordDesigner();
            }
            else
            {
                // We remove the event temporarily so that the TrySaveDirtyLine isn't called

                _chordLinesTreeView.CursorChanged -= OnChordLinesTreeViewCursorChanged;
                _chordLinesTreeView.SetCursor(new TreePath(new int[] {_vm.ActiveLineItem}), _chordLinesTreeView.Columns[0], false);

                RefreshChordLineActions();
                RefreshChordDesigner();

                _chordLinesTreeView.CursorChanged += OnChordLinesTreeViewCursorChanged;
            }

            EnableChordLinesUI(true);
        }

        protected void RefreshChordLineActions()
        {
            EnableChordLineActionsUI(_vm.ActiveLineItem >= 0);
        }

        protected void RefreshChordDesigner()
        {
            bool activeItemIsChord = _vm.ActiveItemIsChord;
            EnableChordDesignerUI(false);
            _titleEntry.Text = activeItemIsChord ? _vm.ChordTitle : "";
            _filenameEntry.Text = activeItemIsChord ? _vm.ChordFileName : "";
            _numStringsSpinButton.Value = activeItemIsChord ? Double.Parse(_vm.ChordNumStrings) : _numStringsSpinButton.Adjustment.Lower;
            _numFretsSpinButton.Value = activeItemIsChord ? Double.Parse(_vm.ChordNumFrets) : _numFretsSpinButton.Adjustment.Lower;
            _baselineSpinButton.Value = activeItemIsChord ? Double.Parse(_vm.ChordBaseLine) : _baselineSpinButton.Adjustment.Lower;
            _barreSpinButton.Value = activeItemIsChord ? Double.Parse(_vm.ChordBarre) : _barreSpinButton.Adjustment.Lower;
            RefreshFretMarkers();
            EnableChordDesignerUI(activeItemIsChord);
            RefreshChordPreview();
        }

        protected void RefreshChordPreview()
        {
            if (!_vm.ActiveItemIsChord)
            {
                _chordPreviewImage.Pixbuf = null;
            }
            else
            {
                try
                {
                    _chordPreviewImage.Pixbuf = GdkUtils.SvgToPixBuf(_vm.ChordSvg);
                }
                catch (Exception)
                {
                    _chordPreviewImage.Pixbuf = new Gdk.Pixbuf("icons/delete_2.png");
                }
            }
        }

        protected void RefreshFretMarkers()
        {
            string marks = _vm.ActiveItemIsChord ? _vm.ChordMarks : "";
            string[] split = marks.Split(';');

            int targetCount = _vm.ActiveItemIsChord ? Int32.Parse(_vm.ChordNumStrings) : 0;

            // Get the right number of markers ready
            while (_marksHBox.Children.Length != targetCount)
            {
                if (_marksHBox.Children.Length < targetCount)
                {
                    // Add new marker
                    Gtk.SpinButton spinButton = new Gtk.SpinButton(-1.0, _numFretsSpinButton.Adjustment.Upper, 1.0);
                    spinButton.Visible = true;
                    spinButton.ValueChanged += OnMarksSpinButtonChanged;
                    _marksHBox.Add(spinButton);
                }
                else
                {
                    // Remove end most marker
                    int endIndex = _marksHBox.Children.Length - 1;

                    Gtk.Widget spinButton = _marksHBox.Children[endIndex];
                    _marksHBox.Remove(spinButton);

                }
            }

            // Get the right values in the spinners
            for (int i = 0; i < _marksHBox.Children.Length; i++)
            {
                Gtk.SpinButton spinButton = (Gtk.SpinButton)_marksHBox.Children[i];
                spinButton.Sensitive = false;
                string s = i < split.Length ? split[i] : "0";
                spinButton.Value = Double.Parse(s);
                spinButton.Sensitive = true;
            }
        }

        protected void EnableChordDesignerUI(bool enable)
        {
            _chordPreviewFrame.Sensitive = enable;
            _chordPropertyFrame.Sensitive = enable;
        }

        protected void EnableChordLinesUI(bool enable)
        {
            _chordLinesTreeViewScrolledWindow.Sensitive = enable;
            _chordLinesToolbar.Sensitive = enable;
        }

        protected void EnableChordLineActionsUI(bool enable)
        {
            MoveLineUpAction.Sensitive = enable;
            MoveLineDownAction.Sensitive = enable;
            EditLineAction.Sensitive = enable && !_vm.ActiveItemIsChord;
            ToggleCommentAction.Sensitive = enable && _vm.ActiveItemCanToggleComment;
            DeleteLineAction.Sensitive = enable;
        }

        #endregion

        protected void OnException(Exception ex)
        {
            Gtk.MessageDialog dialog = new MessageDialog(this, DialogFlags.Modal, MessageType.Error, ButtonsType.None, false, ex.Message);
            dialog.AddButton("Details", ResponseType.Help);
            dialog.AddButton("Close", ResponseType.Close);

            ResponseType result = ResponseType.None;
            do
            {
                result = (ResponseType)dialog.Run();
                if (result == ResponseType.Help)
                {
                    string details = String.IsNullOrEmpty(ex.StackTrace) ? String.Format("{0} thrown.", ex.GetType().ToString()) : String.Format("{0} thrown {1}", ex.GetType().ToString(), ex.StackTrace.Trim());
                    MessageDialog detailsDialog = new MessageDialog(this, DialogFlags.Modal, MessageType.Info, ButtonsType.Close, false, details);
                    detailsDialog.Run();
                    detailsDialog.Destroy();
                }
            }
            while (result != ResponseType.Close);
            dialog.Destroy();
        }

        protected bool ConfirmDestructiveAction(string message)
        {
            MessageDialog dialog = new MessageDialog(this, DialogFlags.Modal, MessageType.Warning, ButtonsType.YesNo, message);

            ResponseType result = (ResponseType)dialog.Run();
            dialog.Destroy();

            return (result == ResponseType.Yes);
        }

        protected void TrySaveDirtyLine()
        {
            if (_vm.ActiveItemIsDirty)
            {
                if (ConfirmDestructiveAction("You've made unsaved changes to the active line. Would you like to save those changes before continuing?"))
                {
                    _vm.SaveLine();
                }
                else
                {
                    _vm.RevertLine();
                }
                RefreshChordLines();
            }
        }

        protected void OnDeleteEvent(object sender, DeleteEventArgs a)
        {
            if (Sensitive && TryClose())
            {
                Application.Quit();
            }
            else
            {
                a.RetVal = true;
            }
        }

        protected void OnChordLinesTreeViewCursorChanged(object sender, System.EventArgs e)
        {
            try
            {
                Gtk.TreePath treePath;
                Gtk.TreeViewColumn treeViewColumn;
                _chordLinesTreeView.GetCursor(out treePath, out treeViewColumn);

                int newItemIndex = treePath.Indices[0];

                TrySaveDirtyLine();

                _vm.ActiveLineItem = newItemIndex;
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnNewActionActivated(object sender, System.EventArgs e)
        {
            if (TryClose())
            {
                NewChordDocument();
            }
        }

        protected void OnOpenActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                if (TryClose())
                {
                    FileChooserDialog dialog = new FileChooserDialog("Open...", this, FileChooserAction.Open,
                        "Cancel", ResponseType.Cancel, "Open", ResponseType.Accept);
        
                    ResponseType result = (ResponseType)dialog.Run();
                    string fileName = dialog.Filename;
                    dialog.Destroy();
        
                    if (result == ResponseType.Accept)
                    {
                        LoadChordDocument(fileName);
                    }
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
        }

        protected void OnSaveActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySave();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnSaveAsActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveAs();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnExportActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();

                string dialogTitle = "";

                if (sender == ExportAsPNGAction)
                {
                    dialogTitle = "Export As PNG...";
                    _vm.SvgWriter = GdkUtils.SvgToPngWriter;
                }
                else if (sender == ExportAsJPGAction)
                {
                    dialogTitle = "Export As JPG...";
                    _vm.SvgWriter = GdkUtils.SvgToJpegWriter;
                }
                else
                {
                    dialogTitle = "Export As SVG...";
                    _vm.SvgWriter = null;
                }

                FileChooserDialog dialog = new FileChooserDialog(dialogTitle, this, FileChooserAction.SelectFolder,
                    "Cancel", ResponseType.Cancel, "Export", ResponseType.Accept);

                ResponseType result = (ResponseType)dialog.Run();
                string outputDir = dialog.Filename;
                dialog.Destroy();

                if (result == ResponseType.Accept)
                {
                    _vm.ExportChords(outputDir);
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                _vm.SvgWriter = null;
                RefreshChordLines();
            }
        }

        protected void OnQuitActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                if (Sensitive && TryClose())
                {
                    Application.Quit();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
        }

        protected void OnAboutActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                AboutDialog dialog = new AboutDialog();
                dialog.ProgramName = com.jonthysell.Chordious.Core.AppInfo.ProgramName;
                dialog.Version = com.jonthysell.Chordious.Core.AppInfo.Version;
                dialog.Copyright = com.jonthysell.Chordious.Core.AppInfo.Copyright;
                dialog.Comments = com.jonthysell.Chordious.Core.AppInfo.Comments;
                dialog.Website = com.jonthysell.Chordious.Core.AppInfo.Website;
                dialog.WebsiteLabel = com.jonthysell.Chordious.Core.AppInfo.WebsiteLabel;
                dialog.Logo = new Gdk.Pixbuf("./icons/logo64.png");
                dialog.License = com.jonthysell.Chordious.Core.AppInfo.License;
                dialog.WrapLicense = true;
                dialog.Run();
                dialog.Destroy();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
        }

        protected void OnAddChordLineActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                _vm.NewChord();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnAddOptionsLineActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                Sensitive = false;

                _vm.NewChordOptions();
                ChordOptionsDialog dialog = new ChordOptionsDialog(_vm);
                ResponseType result = (ResponseType)dialog.Run();
                dialog.Destroy();

                if (result == ResponseType.Accept)
                {
                    _vm.SaveLine();
                }
                else
                {
                    _vm.DeleteLine();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        protected void OnAddCommentLineActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                Sensitive = false;

                _vm.NewComment();
                CommentDialog dialog = new CommentDialog(_vm);
                ResponseType result = (ResponseType)dialog.Run();
                dialog.Destroy();

                if (result == ResponseType.Accept)
                {
                    _vm.SaveLine();
                }
                else
                {
                    _vm.DeleteLine();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        protected void OnMoveLineUpActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                _vm.MoveLineUp();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnMoveLineDownActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                _vm.MoveLineDown();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnEditLineActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                Sensitive = false;

                Dialog dialog = null;
                ResponseType result = ResponseType.None;
                if (_vm.ActiveItemIsComment)
                {
                    dialog = new CommentDialog(_vm);
                }
                else if (_vm.ActiveItemIsChordOptions)
                {
                    dialog = new ChordOptionsDialog(_vm);
                }

                result = (ResponseType)dialog.Run();
                dialog.Destroy();

                if (result == ResponseType.Accept)
                {
                    _vm.SaveLine();
                }
                else
                {
                    TrySaveDirtyLine();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        protected void OnToggleCommentActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                _vm.ToggleComment();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnDeleteLineActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                if (ConfirmDestructiveAction("Are you sure you want to delete this line?"))
                {
                    _vm.DeleteLine();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnChordFinderActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                TrySaveDirtyLine();
                Sensitive = false;

                Dialog dialog = new ChordFinderDialog(_vm);
                dialog.Run();
                dialog.Destroy();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
                Sensitive = true;
            }
        }

        protected void OnRevertChordActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                if (ConfirmDestructiveAction("Are you sure you want to revert this chord? All changes will be lost."))
                {
                    _vm.RevertLine();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordDesigner();
            }
        }

        protected void OnSaveChordActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                _vm.SaveLine();
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
            finally
            {
                RefreshChordLines();
            }
        }

        protected void OnSaveAsDefaultActionActivated(object sender, System.EventArgs e)
        {
            try
            {
                if (ConfirmDestructiveAction("Are you sure you want to set this chord as the default for new chords?"))
                {
                    _vm.SaveLineAsDefault();
                }
            }
            catch (Exception ex)
            {
                OnException(ex);
            }
        }

        protected void OnTitleEntryChanged(object sender, System.EventArgs e)
        {
            if (_titleEntry.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordTitle = _titleEntry.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }

        protected void OnFilenameEntryChanged(object sender, System.EventArgs e)
        {
            if (_filenameEntry.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordFileName = _filenameEntry.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }

        protected void OnNumStringsSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_numStringsSpinButton.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordNumStrings = _numStringsSpinButton.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshFretMarkers();
                    RefreshChordPreview();
                }
            }
        }

        protected void OnNumFretsSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_numFretsSpinButton.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordNumFrets = _numFretsSpinButton.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }

        protected void OnBaselineSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_baselineSpinButton.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordBaseLine = _baselineSpinButton.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }

        protected void OnBarreSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_barreSpinButton.State == StateType.Normal)
            {
                try
                {
                    _vm.ChordBarre = _barreSpinButton.Text;
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }

        protected void OnMarksSpinButtonChanged(object sender, System.EventArgs e)
        {
            if (_marksHBox.State == StateType.Normal)
            {
                try
                {
                    string marks = "";
                    foreach (Gtk.Widget child in _marksHBox.Children)
                    {
                        Gtk.SpinButton fretSpinButton = (Gtk.SpinButton)child;
                        marks += fretSpinButton.Text + ";";
                    }
                    _vm.ChordMarks = marks.TrimEnd(';');
                }
                catch (Exception ex)
                {
                    OnException(ex);
                }
                finally
                {
                    RefreshChordPreview();
                }
            }
        }
    }
}