// 
// MainWindowVM.cs
//  
// Author:
//       Jon Thysell <thysell@gmail.com>
// 
// Copyright (c) 2013 Jon Thysell <http://jonthysell.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using com.jonthysell.Chordious.Core;

namespace com.jonthysell.Chordious.GtkUI
{
    public class MainWindowVM
    {
        public MainWindowVM()
        {
            this._chordDocument = new ChordDocument();
            this.ConfigFile = new ConfigFile("builtin.xml");
            Initialize();
        }

        public MainWindowVM(string inputFile) : this()
        {
            LoadChordDocument(inputFile);
        }

        private void Initialize()
        {
            _activeLineItemIndex = -1;
            _activeLineItem = null;
            ActiveItemIsDirty = false;
            SvgWriter = null;
        }

        #region Global Properties

        public string WindowTitle
        {
            get
            {
                string title = "";
                title += DocumentFileName;
                title += DocumentIsDirty ? "*" : "";
                title += " - Chordious";
                return title;
            }
        }

        public string DocumentFileName
        {
            get
            {
                return String.IsNullOrEmpty(_chordDocument.FileName) ? "Untitled" : _chordDocument.FileName;
            }
        }

        public bool DocumentIsDirty
        {
            get
            {
                return _chordDocument.IsDirty;
            }
        }

        #endregion

        #region Chord Lines

        public List<string> ChordLines
        {
            get
            {
                List<string> chordLines = new List<string>();
                for (int i = 0; i < _chordDocument.Count; i++)
                {
                    chordLines.Add(_chordDocument.GetLine(i).ToString());
                }
                return chordLines;
            }
        }

        public int ActiveLineItem
        {
            get
            {
                return _activeLineItemIndex;
            }
            set
            {
                if (value < 0)
                {
                    _activeLineItem = null;
                    _activeLineItemIndex = -1;
                }
                else
                {
                    _activeLineItem = _chordDocument.GetLine(value);
                    _activeLineItemIndex = value;
                }
                RevertLine();
            }
        }
        private int _activeLineItemIndex;
        private ILineItem _activeLineItem;

        public bool ActiveItemIsChord
        {
            get
            {
                return (null != _activeLineItem && _activeLineItem is Chord);
            }
        }

        public bool ActiveItemIsChordOptions
        {
            get
            {
                return (null != _activeLineItem && _activeLineItem is ChordOptions);
            }
        }

        public bool ActiveItemIsComment
        {
            get
            {
                return (null != _activeLineItem && _activeLineItem is Comment);
            }
        }

        public bool ActiveItemCanToggleComment
        {
            get
            {
                if (ActiveItemIsComment)
                {
                    try
                    {
                        Comment c = (Comment)_activeLineItem;
                        c.GetUncommented();
                    }
                    catch (Exception)
                    {
                        return false;
                    }
                }
                return true;
            }
        }

        public bool ActiveItemIsDirty { get; private set; }

        #endregion

        #region Chord Designer

        public string ChordTitle
        {
            get
            {
                return _chordTitle;
            }
            set
            {
                _chordTitle = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordTitle;

        public string ChordFileName
        {
            get
            {
                return _chordFileName;
            }
            set
            {
                _chordFileName = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordFileName;

        public string ChordNumStrings
        {
            get
            {
                return _chordNumStrings;
            }
            set
            {
                _chordNumStrings = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordNumStrings;

        public string ChordNumFrets
        {
            get
            {
                return _chordNumFrets;
            }
            set
            {
                _chordNumFrets = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordNumFrets;

        public string ChordBaseLine
        {
            get
            {
                return _chordBaseLine;
            }
            set
            {
                _chordBaseLine = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordBaseLine;

        public string ChordBarre
        {
            get
            {
                return _chordBarre;
            }
            set
            {
                _chordBarre = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordBarre;

        public string ChordMarks
        {
            get
            {
                return _chordMarks;
            }
            set
            {
                _chordMarks = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordMarks;

        public string ChordSvg
        {
            get
            {
                Chord c = ChordFromProperties();
                return c.ToSvg(ApplicableChordOptions);
            }
        }

        public string DefaultChordLine
        {
            get
            {
                if (String.IsNullOrEmpty(_defaultChordLine))
                {
                    return "C;4;5;0;-1;0;0;0;3";
                }
                return _defaultChordLine;
            }
            set
            {
                _defaultChordLine = value;
            }
        }
        private string _defaultChordLine;

        #endregion

        #region Comments

        public string CommentText
        {
            get
            {
                return _commentText;
            }
            set
            {
                _commentText = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _commentText;

        #endregion

        #region ChordOptions

        public string ChordOptionsWidth
        {
            get
            {
                return _chordOptionsWidth;
            }
            set
            {
                _chordOptionsWidth = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsWidth;

        public string ChordOptionsHeight
        {
            get
            {
                return _chordOptionsHeight;
            }
            set
            {
                _chordOptionsHeight = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsHeight;

        public string ChordOptionsStrokeWidth
        {
            get
            {
                return _chordOptionsStrokeWidth;
            }
            set
            {
                _chordOptionsStrokeWidth = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsStrokeWidth;

        public string ChordOptionsMargin
        {
            get
            {
                return _chordOptionsMargin;
            }
            set
            {
                _chordOptionsMargin = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsMargin;

        public string ChordOptionsFontSize
        {
            get
            {
                return _chordOptionsFontSize;
            }
            set
            {
                _chordOptionsFontSize = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsFontSize;

        public string ChordOptionsFontFamily
        {
            get
            {
                return _chordOptionsFontFamily;
            }
            set
            {
                _chordOptionsFontFamily = value;
                ActiveItemIsDirty = true;
            }
        }
        private string _chordOptionsFontFamily;

        public string ChordOptionsFontStyle
        {
            get
            {
                return ((FontStyle)ChordOptionsFontStyleIndex).ToString();
            }
            set
            {
                ChordOptionsFontStyleIndex = (int)((FontStyle)Enum.Parse(typeof(FontStyle), value));
            }
        }

        public int ChordOptionsFontStyleIndex
        {
            get
            {
                return _chordOptionsFontStyleIndex;
            }
            set
            {
                _chordOptionsFontStyleIndex = value;
                ActiveItemIsDirty = true;
            }
        }
        private int _chordOptionsFontStyleIndex;

        public string ChordOptionsBarreType
        {
            get
            {
                return ((BarreType)ChordOptionsBarreTypeIndex).ToString();
            }
            set
            {
                ChordOptionsBarreTypeIndex = (int)((BarreType)Enum.Parse(typeof(BarreType), value));
            }
        }

        public int ChordOptionsBarreTypeIndex
        {
            get
            {
                return _chordOptionsBarreTypeIndex;
            }
            set
            {
                _chordOptionsBarreTypeIndex = value;
                ActiveItemIsDirty = true;
            }
        }
        private int _chordOptionsBarreTypeIndex;

        public string ChordOptionsOpenStringType
        {
            get
            {
                return ((OpenStringType)ChordOptionsOpenStringTypeIndex).ToString();
            }
            set
            {
                ChordOptionsOpenStringTypeIndex = (int)((OpenStringType)Enum.Parse(typeof(OpenStringType), value));
            }
        }

        public int ChordOptionsOpenStringTypeIndex
        {
            get
            {
                return _chordOptionsOpenStringTypeIndex;
            }
            set
            {
                _chordOptionsOpenStringTypeIndex = value;
                ActiveItemIsDirty = true;
            }
        }
        private int _chordOptionsOpenStringTypeIndex;

        public bool ChordOptionsFullBarres
        {
            get
            {
                return _chordOptionsFullBarres;
            }
            set
            {
                _chordOptionsFullBarres = value;
                ActiveItemIsDirty = true;
            }
        }
        private bool _chordOptionsFullBarres;


        #endregion

        #region File IO

        public void LoadChordDocument(string inputFile)
        {
            _chordDocument = new ChordDocument(inputFile);
            Initialize();
        }

        public void SaveChordDocument()
        {
            _chordDocument.Save();
        }

        public void SaveChordDocumentAs(string outputFile)
        {
            _chordDocument.SaveAs(outputFile);
        }

        public void ExportActiveChord(string outputDir)
        {
            _chordDocument.ExportChord(_activeLineItemIndex, outputDir, SvgWriter);
        }

        public void ExportChords(string outputDir)
        {
            _chordDocument.ExportChords(outputDir, SvgWriter);
        }

        public ChordDocument.SvgWriter SvgWriter
        {
            get
            {
                return _svgWriter;
            }
            set
            {
                if (null == value)
                {
                    _svgWriter = ChordDocument.DefaultSvgWriter;
                }
                else
                {
                    _svgWriter = value;
                }
            }
        }

        private ChordDocument.SvgWriter _svgWriter;

        #endregion

        #region LineItem Actions

        public void NewChord()
        {
            Chord c = new Chord(DefaultChordLine);
            AddChord(c);
        }

        public void NewComment()
        {
            Comment c = new Comment();
            _chordDocument.AddLine(c);
            ActiveLineItem = _chordDocument.Count - 1;
        }

        public void NewChordOptions()
        {
            ChordOptions co = new ChordOptions();
            _chordDocument.AddLine(co);
            ActiveLineItem = _chordDocument.Count - 1;
        }

        public void MoveLineUp()
        {
            _chordDocument.MoveLineUp(ActiveLineItem);
            ActiveLineItem = Math.Max(ActiveLineItem - 1, 0);
        }

        public void MoveLineDown()
        {
            _chordDocument.MoveLineDown(ActiveLineItem);
            ActiveLineItem = Math.Min(ActiveLineItem + 1, _chordDocument.Count - 1);
        }

        public void ToggleComment()
        {
            _chordDocument.ToggleComment(ActiveLineItem);
            ActiveLineItem = ActiveLineItem;
        }

        public void DeleteLine()
        {
            _chordDocument.DeleteLine(ActiveLineItem);
            int newIndex = ActiveLineItem - 1;
            if (newIndex < 0 && _chordDocument.Count > 0)
            {
                ActiveLineItem = 0;
            }
            else
            {
                ActiveLineItem = newIndex;
            }
        }

        public void RevertLine()
        {
            // ActiveChord
            ChordTitle = (null != ActiveChord) ? ActiveChord.Title : "";
            ChordFileName = (null != ActiveChord && ActiveChord.FileNameSet) ? ActiveChord.FileName : "";
            ChordNumStrings = (null != ActiveChord) ? ActiveChord.NumStrings.ToString() : "";
            ChordNumFrets = (null != ActiveChord) ? ActiveChord.NumFrets.ToString() : "";
            ChordBaseLine = (null != ActiveChord) ? ActiveChord.BaseLine.ToString() : "";
            ChordBarre = (null != ActiveChord) ? ActiveChord.Barre.ToString() : "";

            ChordMarks = "";
            if (null != ActiveChord)
            {
                for (int i = 0; i < ActiveChord.Marks.Length; i++)
                {
                    ChordMarks += ActiveChord.Marks[i].ToString() + ";";
                }
                ChordMarks = ChordMarks.TrimEnd(';');
            }

            // ActiveComment
            CommentText = (null != ActiveComment) ? ActiveComment.Text : "";

            // ActiveChordOptions
            ChordOptionsWidth = (null != ActiveChordOptions) ? ActiveChordOptions.Width.ToString() : "";
            ChordOptionsHeight = (null != ActiveChordOptions) ? ActiveChordOptions.Height.ToString() : "";
            ChordOptionsStrokeWidth = (null != ActiveChordOptions) ? ActiveChordOptions.StrokeWidth.ToString() : "";
            ChordOptionsMargin = (null != ActiveChordOptions) ? ActiveChordOptions.Margin.ToString() : "";
            ChordOptionsFontSize = (null != ActiveChordOptions) ? ActiveChordOptions.FontSize.ToString() : "";
            ChordOptionsFontFamily = (null != ActiveChordOptions) ? ActiveChordOptions.FontFamily.ToString() : "";

            ChordOptionsFontStyleIndex = (null != ActiveChordOptions) ? (int)ActiveChordOptions.FontStyle : 0;
            ChordOptionsBarreTypeIndex = (null != ActiveChordOptions) ? (int)ActiveChordOptions.BarreType : 0;
            ChordOptionsOpenStringTypeIndex = (null != ActiveChordOptions) ? (int)ActiveChordOptions.OpenStringType : 0;
            ChordOptionsFullBarres = (null != ActiveChordOptions) ? (bool)ActiveChordOptions.FullBarres : ChordOptions.DefaultFullBarres;

            ActiveItemIsDirty = false;
        }

        public void SaveLine()
        {
            ILineItem newItem = null;

            if (ActiveItemIsChord)
            {
                newItem = ChordFromProperties();
            }
            else if (ActiveItemIsComment)
            {
                newItem = new Comment(CommentText);
            }
            else if (ActiveItemIsChordOptions)
            {
                newItem = ChordOptionsFromProperties();
            }

            if (null != newItem)
            {
                _chordDocument.UpdateLine(ActiveLineItem, newItem);
            }

            ActiveItemIsDirty = false;
        }

        public void SaveLineAsDefault()
        {
            if (ActiveItemIsChord)
            {
                Chord c = ChordFromProperties();
                DefaultChordLine = c.ToString();
            }
        }

        internal void AddChord(Chord c)
        {
            _chordDocument.AddLine(c);
            ActiveLineItem = _chordDocument.Count - 1;
        }

        #endregion

        #region Active LineItem

        private Chord ActiveChord
        {
            get
            {
                if (ActiveItemIsChord)
                {
                    return (Chord)_activeLineItem;
                }

                return null;
            }
        }

        private Comment ActiveComment
        {
            get
            {
                if (ActiveItemIsComment)
                {
                    return (Comment)_activeLineItem;
                }

                return null;
            }
        }

        private ChordOptions ActiveChordOptions
        {
            get
            {
                if (ActiveItemIsChordOptions)
                {
                    return (ChordOptions)_activeLineItem;
                }

                return null;
            }
        }

        private Chord ChordFromProperties()
        {
            string chordLine = "";

            if (!String.IsNullOrEmpty(ChordFileName))
            {
                chordLine += ChordFileName + ":";
            }

            chordLine += ChordTitle + ";";
            chordLine += String.Format("{0};{1};{2};{3};{4}", ChordNumStrings, ChordNumFrets, ChordBaseLine, ChordBarre, ChordMarks);

            return new Chord(chordLine);
        }

        private ChordOptions ChordOptionsFromProperties()
        {
            string optionsLine = "";

            optionsLine += String.Format("{0};{1};{2};{3};{4};{5};{6};{7};{8};{9}", ChordOptionsWidth, ChordOptionsHeight, ChordOptionsStrokeWidth, ChordOptionsMargin, ChordOptionsFontSize, ChordOptionsFontFamily, ChordOptionsFontStyle, ChordOptionsBarreType, ChordOptionsOpenStringType, ChordOptionsFullBarres);

            return new ChordOptions(optionsLine);
        }

        #endregion

        internal ChordOptions ApplicableChordOptions
        {
            get
            {
                return _chordDocument.GetOptionsForLine(_activeLineItemIndex);
            }
        }

        private ChordDocument _chordDocument;

        internal ConfigFile ConfigFile { get; private set; }
    }
}